import java.lang.Math;
import java.util.ArrayList;
/**
 * This class controls the incoming requests and allocates them to ambulances.
 * A request is called and this class checks every station for a free ambulance.
 * If no ambualance is found then it places the patient on a waiting list.
 * As soon a ambulance becomes free it will be assigned to a patient in the
 * waiting list.
 *
 * @author                  Dylan McKendry
 */
public class Dispatcher
{
    private ArrayList<AmbulanceStation> stations;
    private ArrayList<Patient> waiting;
    private int totalStations = 0;
    private boolean[][] cityRoad;

    public Dispatcher(boolean[][] cityRoad)
    {
        this.cityRoad = cityRoad;
        stations = new ArrayList<AmbulanceStation>();
        waiting = new ArrayList<Patient>();
    }

    public void simulate()
    {
        if (!waiting.isEmpty()) {
            Patient patient = waiting.get(0);
            int x = patient.getX();
            int y = patient.getY();
            Ambulance ambulance = findAmbulance(x, y);

            if (ambulance != null) {
                ambulance.pickUp(patient);
                waiting.remove(0);
            }
        }

        for (AmbulanceStation s : stations) {
            s.simulate();
        }
    }


    public ArrayList<Ambulance> getAllAmbulances()
    {
        ArrayList<Ambulance> all = new ArrayList<Ambulance>();
        for (AmbulanceStation s : stations) {
            all.addAll(all.size(), s.getAmbulances());
        }
        return all;
    }

    public void newStation(int x, int y)
    {
        stations.add(new AmbulanceStation(x, y, cityRoad));
        totalStations++;
    }


    public void pickUpRequest(Patient patient)
    {
        int x = patient.getX();
        int y = patient.getY();
        Ambulance ambulance = findAmbulance(x, y);

        if (ambulance == null) {
            addToWaiting(patient);
        }
        else {
            ambulance.pickUp(patient);
        }
    }


    public void removePatient(Patient patient)
    {
        int patientID = patient.getID();
        for (Patient p : waiting) {
            if (p.getID() == patientID) {
                waiting.remove(patientID);
            }
        }
    }


   /*
    * Adds a patient to the waiting list.
    * So that the first element has lowset life
    */
    private void addToWaiting(Patient patient)
    {
        boolean inserted = false;

        if (waiting.isEmpty()) {
            waiting.add(patient);
            return;
        }

        int size = waiting.size();

        for (int i = 0; i < size; i++) {
            Patient current = waiting.get(i);
            if (patient.getLife() < current.getLife()) {
                waiting.add(i, patient);
                inserted = true;
            }
        }

        if (inserted == false) {
            waiting.add(patient);
        }
    }


   /*
    * Finds the closest available ambulance to the position of the request.
    * If none are available then returns null.
    */
    private Ambulance findAmbulance(int x, int y)
    {
        ArrayList<Ambulance> available = findFreeAmbulances();

        if (available.size() != 0) {
            Ambulance ambulance = findClosestAmbulance(available, x, y);
            return ambulance;
        }
        else {
            return null;
        }
    }


   /*
    * Finds a ArrayList of all availble ambulances from every station.
    * If none are available then returns null.
    */
    private ArrayList<Ambulance> findFreeAmbulances()
    {
        ArrayList<Ambulance> available = new ArrayList<Ambulance>();

        for (AmbulanceStation current : stations) {
            ArrayList<Ambulance> currentAmbulances = current.freeAmbulances();
            if (currentAmbulances.size() != 0)
            {
                available.addAll(currentAmbulances);
            }
        }
        return available;
    }


   /*
    * Finds the closest ambulance to the position of the request.
    * If none are available then returns null.
    */
    private Ambulance findClosestAmbulance(ArrayList<Ambulance> available, int x, int y)
    {
        Ambulance ambulance = available.get(0);
        double distance = findDistance(x, y, ambulance.getX(),
                                       ambulance.getY());

        for (int i = 1; i < available.size(); i++) {
            Ambulance current = available.get(i);
            double otherDistance = findDistance(x, y, current.getX(),
                                             current.getY());
            if (otherDistance < distance) {
                ambulance = current;
                distance = otherDistance;
            }
        }
        return ambulance;
    }



   /*
    * Finds the distance from two points A(xA, yA) and B(xB, yB).
    */
    private double findDistance(int xA, int yA, int xB, int yB)
    {
        double xDistance = xA - xB;
        double yDistance = yA - yB;
        double distance = Math.hypot(xDistance, yDistance);
        return distance;
    }

}