import java.util.ArrayList;

/**
 * Finds the shortest path between two points.
 * The path is found by attempting possible paths in the correct direction.
 * The various paths are stored in and arraylist of characters.
 *
 * @author                  Dylan McKendry
 */
 
public class PathFinder
{
    private int startX;
    private int startY;
    private int endX;
    private int endY;
    
    // used as the max direction the path can go
    private int maxX;
    private int maxY;

    private boolean[][] cityRoad;
    private boolean[][] copyCityRoad;

    private ArrayList<Path> paths;

    private Path finalPath;

    private boolean right;
    private boolean left;
    private boolean up;
    private boolean down;

   /**
    * The constructor for PathFinder.
    * It loads the city map into the object.
    *
    * @param  cityRoad      a 2-d boolean array of the city, true is a road
    */
    public PathFinder(boolean[][] cityRoad)
    {
        copyCityRoad = cityRoad;
        paths = new ArrayList<Path>();

        maxX = cityRoad.length - 1;
        maxY = cityRoad[0].length - 1;
    }
    

   /**
    * Inputs the start and end coordinates.
    * The coordinates are loaded then it calls another class to find the path.
    * It returns characters U, D, L, R.
    * Which refer to up, down, left, right.
    *
    * @param  startX        the starting x coordinate
    * @param  startY        the starting y coordinate
    * @param  endX          the ending x coordinate
    * @param  endY          the ending y coordinate
    * @return               an arraylist of characters(U,D,L,R) or null
    */
    public String findPath(int startX, int startY, int endX, int endY)
    {

        cityRoad = new boolean[copyCityRoad.length][];
        for (int i = 0; i < copyCityRoad.length; i++) {
            cityRoad[i] = (boolean[])copyCityRoad[i].clone();
        }

        this.startX = startX;
        this.startY = startY;
        this.endX = endX;
        this.endY = endY;

        cityRoad[startX][startY] = false;
        
        paths.clear();
        finalPath = null;

        paths.add(new Path(startX, startY));
        findPaths();

        if(finalPath != null)
        {
            return finalPath.getDirections();
        }
        return null;
    }

   /*
    * Contains the loop to find the quickest path
    */
    private void findPaths()
    {
        boolean finished = false;
        boolean noPath = false;
        
        while (!finished && !noPath) {
            int numberOfPaths = paths.size();
            finished = false;
            noPath = true;  //make sure there is an active path
            for (int i = 0; i < numberOfPaths; i++) {
                Path path = paths.get(i);
                if (!path.isDeadEnd()) {
                    noPath = false;
                    finished = findPossibleDirections(i);
                }
                if(finished)
                {
                    break;
                }
            }

        }
    }
 

   /*
    * Trys every direction to test if a path could go that way.
    * It will then continue the path in every direction.
    */
    private boolean findPossibleDirections(int i)
    {
        Path path = paths.get(i);

        int currentX = path.getX();
        int currentY = path.getY();
        
        right = false;
        left = false;
        up = false;
        down = false;

        
        int difX = currentX - endX;
        int difY = currentY - endY;
        
        if ((difX == 0) && (difY == 0)) {
            finalPath = path;
            return true;
        }


        int count = 0;
        
        if (currentX < maxX) {
            //try right
            if(cityRoad[currentX+1][currentY])
            {
                count++;
                right = true;
            }
        }

        if(currentX > 0) {
            //try left
            if(cityRoad[currentX-1][currentY])
            {
                count++;
                left = true;
            }
        }

        if(currentY > 0) {
            //try up
            if(cityRoad[currentX][currentY-1])
            {
                count++;
                up = true;
            }
        }
        
        if (currentY < maxY) {
            //try down
            if(cityRoad[currentX][currentY+1])
            {
                count++;
                down = true;
            }
        }
        
        //depending how many paths continue a path for each
        switch (count) {
            case 4:
                Path p4 = new Path(path.getDirections(), currentX, currentY);
                paths.add(p4);
                updatePath(p4);
            case 3:
                Path p3 = new Path(path.getDirections(), currentX, currentY);
                paths.add(p3);
                updatePath(p3);
            case 2:
                Path p2 = new Path(path.getDirections(), currentX, currentY);
                paths.add(p2);
                updatePath(p2);
            case 1:
                updatePath(path);
                break;
            default:
                path.deadEnd();
                break;
        }

        return false;

    }

   /*
    * Updates a path in a direction.
    */
    private void updatePath(Path path)
    {
        boolean updated = false;

        if(right) {
            path.addChar('R');
            right = false;
            updated = true;
        }
        else if(left && !updated) {
            path.addChar('L');
            left = false;
            updated = true;
        }
        else if(down && !updated) {
            path.addChar('D');
            down = false;
            updated = true;
        }
        else if(up && !updated) {
            path.addChar('U');
            up = false;
        }
        
        //sets the path to false so nothing trys it again
        cityRoad[path.getX()][path.getY()] = false;
    }
}
